@@ -17,10 +17,14 @@ use XSLoader;
 use base qw/Devel::Declare::Context::Simple/;
 
 
-our $VERSION = '1.001001';
+our $VERSION = '1.002000';
 
 # These are private state variables. Mess with them at your peril
 our ($CHECK_OP_HOOK, $CHECK_OP_DEPTH) = (undef, 0);
+our $NEXT_EVAL_IS_TRY;
+
+# Stack of state for tacking nested. Each value is number of catch blocks at
+# the current level. We are nested if @STATE > 1
 our (@STATE);
 
 XSLoader::load(__PACKAGE__, $VERSION);
@@ -33,11 +37,15 @@ use Sub::Exporter -setup => {
   installer => sub {
     my ($args, $to_export) = @_;
     my $pack = $args->{into};
+    my $ctx_class = $args->{parser} || 'TryCatch';
+
+    TryCatch::XS::install_try_op_check();
+
     foreach my $name (@$to_export) {
       if (my $parser = __PACKAGE__->can("_parse_${name}")) {
         Devel::Declare->setup_for(
           $pack,
-          { $name => { const => sub { $parser->($pack, @_) } } },
+          { $name => { const => sub { $ctx_class->$parser($pack, @_) } } },
         );
       }
     }
@@ -49,7 +57,6 @@ use Sub::Exporter -setup => {
 
 # The actual try call itself. Nothing to do with parsing.
 sub try () {
-  warn "non-shadowed - shouldn't happen";
   return;
 }
 
@@ -69,12 +76,12 @@ sub check_tc {
 
 # Replace 'try {' with an 'try; { local $@; eval {'
 sub _parse_try {
-  my $pack = shift;
+  my ($class,$pack, @args) = @_;
 
   # Hide Devel::Declare from carp;
   local $Carp::Internal{'Devel::Declare'} = 1;
 
-  my $ctx = TryCatch->new->init(@_);
+  my $ctx = $class->new->init(@args);
 
   $ctx->skip_declarator;
   $ctx->skipspace;
@@ -88,28 +95,30 @@ sub _parse_try {
   $ctx->shadow(sub () { } );
 
   $ctx->inject_if_block(
-    $ctx->injected_try_code . $ctx->scope_injector_call,
+    $ctx->injected_try_code . $ctx->scope_injector_call('{}'),
     q#;#
   ) or croak "block required after try";
 
   #$ctx->debug_linestr("try");
-  if (! $CHECK_OP_DEPTH++) {
-    $CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
+  if (! $TryCatch::CHECK_OP_DEPTH) {
+    $TryCatch::CHECK_OP_DEPTH++;
+    $TryCatch::CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
   }
 
   $ctx->debug_linestr('post try');
 
   # Number of catch blocks found, we only care about 0,1 and +1 cases tho
-  push @STATE, 0;
+  $ctx->state_new_block();
   
 }
 
+
 sub injected_try_code {
   # try { ...
   # ->
   # try; { local $@; eval { ...
 
-  return @STATE > 1
+  return $_[0]->state_is_nested()
        ? 'local $TryCatch::CTX = Scope::Upper::HERE; eval {' # Nested case
        : 'local $@; eval {'
 }
@@ -120,12 +129,36 @@ sub injected_after_try {
 }
 
 sub injected_no_catch_code {
-  return "};";
+  # This undef is to ensure that there is the eval{}; is called in void context
+  # i.e that its not the last op in a subroutine
+  return "};undef;";
 }
 
+sub injected_post_catch_code {
+  # We do it like this so that PROPGATE gets called, in case anyone is using it
+  return 'else { $@ = $TryCatch::Error; die } };undef;';
+}
+
+
+sub scope_injector_call {
+  my $self = shift;
+  my $inject = shift;
+  return ' BEGIN { ' . ref($self) . "->inject_scope(${inject}) }; ";
+}
+
+
 sub inject_scope {
+  my ($class, $opts) = @_;
+
+  if ($opts->{hook}) {
+    if (! $TryCatch::CHECK_OP_DEPTH) {
+      $TryCatch::CHECK_OP_DEPTH++;
+      $TryCatch::CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
+    }
+  }
+
   on_scope_end { 
-    block_postlude() 
+    $class->block_postlude() 
   }
 }
 
@@ -133,7 +166,8 @@ sub inject_scope {
 # Look ahead and determine what action to take based on wether or note we
 # see a 'catch' token after the block
 sub block_postlude {
-  my $ctx = TryCatch->new->init(
+  my ($class) = @_;
+  my $ctx = $class->new->init(
     '', 
     Devel::Declare::get_linestr_offset()
   );
@@ -151,8 +185,9 @@ sub block_postlude {
     $ctx->{Declarator} = $toke;
   }
 
-  if ($CHECK_OP_DEPTH && --$CHECK_OP_DEPTH == 0) {
-    TryCatch::XS::uninstall_return_op_check($CHECK_OP_HOOK);
+  if ($TryCatch::CHECK_OP_DEPTH && --$TryCatch::CHECK_OP_DEPTH == 0) {
+    TryCatch::XS::uninstall_return_op_check($TryCatch::CHECK_OP_HOOK);
+    $TryCatch::CHECK_OP_HOOK = '';
   }
 
   if ($toke eq 'catch') {
@@ -162,9 +197,16 @@ sub block_postlude {
     $ctx->_parse_catch;
 
   } else  {
-    my $code = $STATE[-1] == 0
-             ? $ctx->injected_no_catch_code
-             : '}';
+
+    # No (more) catch blocks, so write the postlude
+    my $code;
+    if ($ctx->state_have_catch_block) {
+      $code = $ctx->injected_post_catch_code;
+    }
+    else {
+      $code = $ctx->injected_no_catch_code;
+      $TryCatch::NEXT_EVAL_IS_TRY = 1;
+    }
 
     substr($linestr, $offset, 0, $code);
 
@@ -173,7 +215,7 @@ sub block_postlude {
     $ctx->debug_linestr("finalizer");
 
     # This try/catch stmt is finished
-    pop @STATE;
+    $ctx->state_end_block();
   }
 }
 
@@ -181,100 +223,148 @@ sub block_postlude {
 # turn 'catch() {' into '->catch({ TC_check_code;'
 # the '->' is added by one of the postlude hooks
 sub _parse_catch {
-  my $ctx = shift;
-  my $pack = $ctx->get_curstash_name;
-
+  my ($ctx) = @_;
 
   # Hide Devel::Declare from carp;
   local $Carp::Internal{'Devel::Declare'} = 1;
   local $Carp::Internal{'B::Hooks::EndOfScope'} = 1;
   local $Carp::Internal{'TryCatch'} = 1;
+  local $Carp::Internal{'TryCatch::Basic'} = 1;
 
-  # This isn't a normal DD-callback, so we can strip_name to get rid of try
+  # This isn't a normal DD-callback, so we can strip_name to get rid of 'catch'
   my $offset = $ctx->offset;
   $ctx->strip_name;
   $ctx->skipspace;
-  my $new_offset = $ctx->offset;
  
   $ctx->debug_linestr('catch');
   my $linestr = $ctx->get_linestr;
 
-  my $code;
-  my $var_code = "";
+  my ($code, $var_code, @conditions) = ("","");
 
-  my @conditions;
   # optional ()
   if (substr($linestr, $ctx->offset, 1) eq '(') {
-    my $proto = $ctx->strip_proto;
-    croak "Run-away catch signature"
-      unless (length $proto);
-    
-    my $sig = Parse::Method::Signatures->new(
-      input => $proto,
-      from_namespace => $pack,
-    );
-    my $errctx = $sig->ppi;
-    my $param = $sig->param;
-
-    $sig->error( $errctx, "Parameter expected")
-      unless $param;
-
-    my $left = $sig->remaining_input;
-
-    croak "TryCatch can't handle un-named vars in catch signature" 
-      unless $param->can('variable_name');
-
-    my $name = $param->variable_name;
-    $var_code = "my $name = \$TryCatch::Error;";
-
-    # (TC $var)
-    if ($param->has_type_constraints) {
-      my $tc = $param->meta_type_constraint;
-      $TC_LIBRARY->{"$tc"} = $tc;
-      push @conditions, "TryCatch->check_tc('$tc')";
-    }
-
-    # ($var where { $_ } )
-    if ($param->has_constraints) {
-      foreach my $con (@{$param->constraints}) {
-        $con =~ s/^{|}$//g;
-        push @conditions, "do {local \$_ = \$TryCatch::Error; $con }";
-      }
-    }
-
-    $linestr = $ctx->get_linestr;
+    ($var_code, @conditions) = $ctx->parse_proto()
   }
 
-  $code = $ctx->injected_after_try
-    if $STATE[-1] == 0;
+  @conditions = ('1') unless @conditions;
 
-  @conditions = ('1')
-    unless @conditions;
+  unless ($ctx->state_have_catch_block()) {
+    $TryCatch::NEXT_EVAL_IS_TRY = 1;
+    $code = $ctx->injected_after_try
+          . "if (";
+  }
+  else {
+    $code = "elsif (";
+  }
 
-  $code .= $STATE[-1] < 1
-         ? "if ("
-         : "elsif (";
+  $var_code = $ctx->scope_injector_call( 
+    $ctx->state_is_nested() ? '{hook => 1}' : '{}'
+  ) . $var_code;
 
   $ctx->inject_if_block(
-    $ctx->scope_injector_call . $var_code,
+    $var_code,
     $code . join(' && ', @conditions) . ')'
   ) or croak "block required after catch";
 
   $ctx->debug_linestr('post catch');
 
-  #if (! $CHECK_OP_DEPTH++) {
-  #  $CHECK_OP_HOOK = TryCatch::XS::install_return_op_check();
-  #}
-
-  $STATE[-1]++;
+  $ctx->state_parsed_catch();
 }
 
-sub debug_linestr {
+sub parse_proto {
+  my ($self) = @_;
+
+  my $proto = $self->strip_proto;
+  croak "Run-away catch signature"
+    unless (length $proto);
+
+  return $self->parse_proto_using_pms($proto);
+}
+
+sub _string_to_tc {
+  my ($class, $name) = @_;
+
+  my $tc = $class->find_registered_constraint($name);
+
+  return $tc if ref $tc;
+
+  return Moose::Util::TypeConstraints::find_or_create_isa_type_constraint($name)
+}
+
+sub parse_proto_using_pms {
+  my ($self, $proto) = @_;
+
+  my @conditions;
+
+  my $sig = Parse::Method::Signatures->new(
+    input => $proto,
+    from_namespace => $self->get_curstash_name,
+    type_constraint_callback => \&_string_to_tc,
+  );
+  my $errctx = $sig->ppi;
+  my $param = $sig->param;
+
+  $sig->error( $errctx, "Parameter expected")
+    unless $param;
+
+  my $left = $sig->remaining_input;
+
+  my $var_code = '';
+
+  if (my $var_name = $param->can('variable_name') ) {
+
+    my $name = $param->$var_name();
+    $var_code = "my $name = \$TryCatch::Error;";
+  }
+
+  # (TC $var)
+  if ($param->has_type_constraints) {
+    my $tc = $param->meta_type_constraint;
+    $TryCatch::TC_LIBRARY->{"$tc"} = $tc;
+    push @conditions, "TryCatch->check_tc('$tc')";
+  }
+
+  # ($var where { $_ } )
+  if ($param->has_constraints) {
+    foreach my $con (@{$param->constraints}) {
+      $con =~ s/^{|}$//g;
+      push @conditions, "do {local \$_ = \$TryCatch::Error; $con }";
+    }
+  }
+
+  return $var_code, @conditions;
+}
+
+# Functions that wrap @STATE maniuplation into useful names
+sub state_new_block {
+  push @TryCatch::STATE, 0;
+  return
+}
+
+sub state_end_block {
+  pop @TryCatch::STATE;
+  return;
+}
+
+sub state_is_nested {
+  return @TryCatch::STATE > 1;
+}
+
+sub state_have_catch_block {
+  return $TryCatch::STATE[-1] > 0;
+}
+
+sub state_parsed_catch {
+  $TryCatch::STATE[-1]++;
+}
+
+*debug_linestr = !( ($ENV{TRYCATCH_DEBUG} || 0) & 1)
+               ? sub {}
+               : sub {
   my ($ctx, $message) = @_;
 
-  return unless $ENV{TRYCATCH_DEBUG};
- 
   local $Carp::Internal{'TryCatch'} = 1;
+  local $Carp::Internal{'TryCatch::Basic'} = 1;
   local $Carp::Internal{'Devel::Declare'} = 1;
   local $Carp::Internal{'B::Hooks::EndOfScope'} = 1;
   local $Carp::Internal{'Devel::PartialDump'} = 1;
@@ -284,7 +374,7 @@ sub debug_linestr {
 
   warn   "  Substr: ", Devel::PartialDump::dump(substr($ctx->get_linestr, $ctx->offset)),
        "\n  Whole:  ", Devel::PartialDump::dump($ctx->get_linestr), "\n\n";
-}
+};
 
 
 1;
@@ -295,6 +385,14 @@ __END__
 
 TryCatch - first class try catch semantics for Perl, without source filters.
 
+=head1 DESCRIPTION
+
+This module aims to provide a nicer syntax and method to catch errors in Perl,
+similar to what is found in other languages (such as Java, Python or C++).  The
+standard method of using C<< eval {}; if ($@) {} >> is often prone to subtle
+bugs, primarily that its far too easy to stomp on the error in error handlers.
+And also eval/if isn't the nicest idiom.
+
 =head1 SYNOPSIS
 
  use TryCatch;
@@ -380,22 +478,6 @@ Write some more documentation
 
 =back
 
-=head1 KNOWN BUGS
-
-Currently C<@_> is not accessible inside try or catch blocks, so assign this to
-a lexical variable outside if you wish to access function arguments. i.e.: 
-
- sub foo {
-   try { return $_[0] };
- }
-
-will not work, instead you must do something similar to this:
-
- sub foo {
-   my ($foo) = @_;
-   try { return $foo }
- }
-
 =head1 SEE ALSO
 
 L<MooseX::Types>, L<Moose::Util::TypeConstraints>, L<Parse::Method::Signatures>.
